.sd-card {
  background-color: var(--pst-color-on-background);
  border-radius: 10px;
  padding: 30px 10px 20px 10px;
  margin: 10px 0px;
}

.sd-card .sd-card-header .sd-card-text {
  margin: 0px;
}

.sd-card .sd-card-header {
  border: none;
  text-align: center;
  font-size: var(--pst-font-size-h4);
  font-weight: bold;
  padding: 0.5rem 0rem 0.5rem 0rem;
}

.sd-card .sd-card-footer {
  border: none;
}

.sd-card .sd-card-footer .sd-card-text {
  max-width: 220px;
  margin-left: auto;
  margin-right: auto;
}

.sd-card .sd-btn {
  border-radius: 20px;
}

.animated-sd-card {
  border-radius: 10px;
  padding: 30px 10px 20px 10px;
  margin: 10px 0px;
}

.animated-sd-card:hover {
  border-color: var(--pst-color-secondary);
  transform: scale(1.05);
  -webkit-transition: all .2s;
  -moz-transition: all .2s;
  transition: all .2s;
}

.animated-sd-card .sd-btn:hover {
  -webkit-animation: pulse 2s infinite;
  animation: pulse512 1.5s infinite;
  background-size: 200% auto;
}
